@@ -40,7 +40,7 @@ try:
 except ImportError:
     from pipes import quote
 
-from devlib.exception import TargetTransientError, TargetStableError, HostError
+from devlib.exception import TargetTransientError, TargetStableError, HostError, TargetTransientCalledProcessError, TargetStableCalledProcessError
 from devlib.utils.misc import check_output, which, ABI_MAP, redirect_streams, get_subprocess
 from devlib.connection import ConnectionBase, AdbBackgroundCommand, PopenBackgroundCommand, PopenTransferManager
 
@@ -315,6 +315,7 @@ class AdbConnection(ConnectionBase):
         return self._push_pull('pull', sources, dest, timeout)
 
     def _push_pull(self, action, sources, dest, timeout):
+        sources = list(sources)
         paths = sources + [dest]
 
         # Quote twice to avoid expansion by host shell, then ADB globbing
@@ -332,9 +333,19 @@ class AdbConnection(ConnectionBase):
     # pylint: disable=unused-argument
     def execute(self, command, timeout=None, check_exit_code=False,
                 as_root=False, strip_colors=True, will_succeed=False):
+        if as_root and self.connected_as_root:
+            as_root = False
         try:
             return adb_shell(self.device, command, timeout, check_exit_code,
                              as_root, adb_server=self.adb_server, su_cmd=self.su_cmd)
+        except subprocess.CalledProcessError as e:
+            cls = TargetTransientCalledProcessError if will_succeed else TargetStableCalledProcessError
+            raise cls(
+                e.returncode,
+                command,
+                e.output,
+                e.stderr,
+            )
         except TargetStableError as e:
             if will_succeed:
                 raise TargetTransientError(e)
@@ -342,6 +353,8 @@ class AdbConnection(ConnectionBase):
                 raise
 
     def background(self, command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, as_root=False):
+        if as_root and self.connected_as_root:
+            as_root = False
         bg_cmd = self._background(command, stdout, stderr, as_root)
         self._current_bg_cmds.add(bg_cmd)
         return bg_cmd
@@ -404,6 +417,9 @@ class AdbConnection(ConnectionBase):
         logger.debug("ls command is set to {}".format(self.ls_command))
 
     def _setup_su(self):
+        # Already root, nothing to do
+        if self.connected_as_root:
+            return
         try:
             # Try the new style of invoking `su`
             self.execute('ls', timeout=self.timeout, as_root=True,
@@ -558,10 +574,15 @@ def adb_shell(device, command, timeout=None, check_exit_code=False,
         exit_code = exit_code.strip()
         re_search = AM_START_ERROR.findall(output)
         if exit_code.isdigit():
-            if int(exit_code):
-                message = ('Got exit code {}\nfrom target command: {}\n'
-                           'OUTPUT: {}\nSTDERR: {}\n')
-                raise TargetStableError(message.format(exit_code, command, output, error))
+            exit_code = int(exit_code)
+            if exit_code:
+                raise subprocess.CalledProcessError(
+                    exit_code,
+                    command,
+                    output,
+                    error,
+                )
+
             elif re_search:
                 message = 'Could not start activity; got the following:\n{}'
                 raise TargetStableError(message.format(re_search[0]))
@@ -575,7 +596,7 @@ def adb_shell(device, command, timeout=None, check_exit_code=False,
                           '-----\nSTDERR:\n-----\n{}\n-----'
                 raise TargetTransientError(message.format(raw_output, error))
 
-    return output + error
+    return '\n'.join(x for x in (output, error) if x)
 
 
 def adb_background_shell(conn, command,
@@ -600,7 +621,7 @@ def adb_background_shell(conn, command,
     adb_cmd = get_adb_command(device, 'shell', adb_server)
     full_command = '{} {}'.format(adb_cmd, quote(command))
     logger.debug(full_command)
-    p = subprocess.Popen(full_command, stdout=stdout, stderr=stderr, shell=True)
+    p = subprocess.Popen(full_command, stdout=stdout, stderr=stderr, stdin=subprocess.PIPE, shell=True)
 
     # Out of band PID lookup, to avoid conflicting needs with stdout redirection
     find_pid = '{} ps -A -o pid,args | grep {}'.format(conn.busybox, quote(uuid_var))
