@@ -31,6 +31,7 @@ import contextlib
 import weakref
 import select
 import copy
+import functools
 from pipes import quote
 from future.utils import raise_from
 
@@ -51,7 +52,10 @@ from pexpect import EOF, TIMEOUT, spawn
 
 # pylint: disable=redefined-builtin,wrong-import-position
 from devlib.exception import (HostError, TargetStableError, TargetNotRespondingError,
-                              TimeoutError, TargetTransientError)
+                              TimeoutError, TargetTransientError,
+                              TargetCalledProcessError,
+                              TargetTransientCalledProcessError,
+                              TargetStableCalledProcessError)
 from devlib.utils.misc import (which, strip_bash_colors, check_output,
                                sanitize_cmd_template, memoized, redirect_streams)
 from devlib.utils.types import boolean
@@ -303,65 +307,8 @@ class SshConnectionBase(ConnectionBase):
         self.sudo_cmd = sanitize_cmd_template(sudo_cmd)
         self.platform = platform
         self.strict_host_check = strict_host_check
-        self.options = {}
         logger.debug('Logging in {}@{}'.format(username, host))
 
-    def fmt_remote_path(self, path):
-        return '{}@{}:{}'.format(self.username, self.host, path)
-
-    def push(self, sources, dest, timeout=30):
-        # Quote the destination as SCP would apply globbing too
-        dest = self.fmt_remote_path(quote(dest))
-        paths = sources + [dest]
-        return self._scp(paths, timeout)
-
-    def pull(self, sources, dest, timeout=30):
-        # First level of escaping for the remote shell
-        sources = ' '.join(map(quote, sources))
-        # All the sources are merged into one scp parameter
-        sources = self.fmt_remote_path(sources)
-        paths = [sources, dest]
-        self._scp(paths, timeout)
-
-    def _scp(self, paths, timeout=30):
-        # NOTE: the version of scp in Ubuntu 12.04 occasionally (and bizarrely)
-        # fails to connect to a device if port is explicitly specified using -P
-        # option, even if it is the default port, 22. To minimize this problem,
-        # only specify -P for scp if the port is *not* the default.
-        port_string = '-P {}'.format(quote(str(self.port))) if (self.port and self.port != 22) else ''
-        keyfile_string = '-i {}'.format(quote(self.keyfile)) if self.keyfile else ''
-        options = " ".join(["-o {}={}".format(key, val)
-                            for key, val in self.options.items()])
-        paths = ' '.join(map(quote, paths))
-        command = '{} {} -r {} {} {}'.format(scp,
-                                                options,
-                                                keyfile_string,
-                                                port_string,
-                                                paths)
-        command_redacted = command
-        logger.debug(command)
-        if self.password:
-            command, command_redacted = _give_password(self.password, command)
-        try:
-            check_output(command, timeout=timeout, shell=True)
-        except subprocess.CalledProcessError as e:
-            raise_from(HostError("Failed to copy file with '{}'. Output:\n{}".format(
-                command_redacted, e.output)), None)
-        except TimeoutError as e:
-            raise TimeoutError(command_redacted, e.output)
-
-    def _get_default_options(self):
-        if self.strict_host_check:
-            options = {
-                'StrictHostKeyChecking': 'yes',
-            }
-        else:
-            options = {
-                'StrictHostKeyChecking': 'no',
-                'UserKnownHostsFile': '/dev/null',
-            }
-        return options
-
 
 class SshConnection(SshConnectionBase):
     # pylint: disable=unused-argument,super-init-not-called
@@ -405,8 +352,6 @@ class SshConnection(SshConnectionBase):
 
         if self.use_scp:
             logger.debug('Using SCP for file transfer')
-            _check_env()
-            self.options = self._get_default_options()
         else:
             logger.debug('Using SFTP for file transfer')
 
@@ -465,35 +410,31 @@ class SshConnection(SshConnectionBase):
     def _get_progress_cb(self):
         return self.transfer_mgr.progress_cb if self.transfer_mgr is not None else None
 
+    # Limit the number of opened channels to a low number, since some servers
+    # will reject more connections request. For OpenSSH, this is controlled by
+    # the MaxSessions config.
+    @functools.lru_cache(maxsize=1)
+    def _cached_get_sftp(self):
+        try:
+            sftp = self.client.open_sftp()
+        except paramiko.ssh_exception.SSHException as e:
+            if 'EOF during negotiation' in str(e):
+                raise TargetStableError('The SSH server does not support SFTP. Please install and enable appropriate module.') from e
+            else:
+                raise
+        return sftp
+
     def _get_sftp(self, timeout):
-        sftp = self.client.open_sftp()
+        sftp = self._cached_get_sftp()
         sftp.get_channel().settimeout(timeout)
         return sftp
 
+    @functools.lru_cache()
     def _get_scp(self, timeout):
         return SCPClient(self.client.get_transport(), socket_timeout=timeout, progress=self._get_progress_cb())
 
     def _push_file(self, sftp, src, dst):
-        try:
-            sftp.put(src, dst, callback=self._get_progress_cb())
-        # Maybe the dst was a folder
-        except OSError as orig_excep:
-            # If dst was an existing folder, we add the src basename to create
-            # a new destination for the file as cp would do
-            new_dst = os.path.join(
-                dst,
-                os.path.basename(src),
-            )
-            logger.debug('Trying: {} -> {}'.format(src, new_dst))
-            try:
-                sftp.put(src, new_dst, callback=self._get_progress_cb())
-            # This still failed, which either means:
-            # * There are some missing folders in the dirnames
-            # * Something else SFTP-related is wrong
-            except OSError as e:
-                # Raise the original exception, as it is closer to what the
-                # user asked in the first place
-                raise orig_excep
+        sftp.put(src, dst, callback=self._get_progress_cb())
 
     @classmethod
     def _path_exists(cls, sftp, path):
@@ -505,29 +446,13 @@ class SshConnection(SshConnectionBase):
             return True
 
     def _push_folder(self, sftp, src, dst):
-        # Behave like the "mv" command or adb push: a new folder is created
-        # inside the destination folder, rather than merging the trees, but
-        # only if the destination already exists. Otherwise, it is use as-is as
-        # the new hierarchy name.
-        if self._path_exists(sftp, dst):
-            dst = os.path.join(
-                dst,
-                os.path.basename(os.path.normpath(src)),
-            )
-
-        return self._push_folder_internal(sftp, src, dst)
-
-    def _push_folder_internal(self, sftp, src, dst):
-        # This might fail if the folder already exists
-        with contextlib.suppress(IOError):
-            sftp.mkdir(dst)
-
+        sftp.mkdir(dst)
         for entry in os.scandir(src):
             name = entry.name
             src_path = os.path.join(src, name)
             dst_path = os.path.join(dst, name)
             if entry.is_dir():
-                push = self._push_folder_internal
+                push = self._push_folder
             else:
                 push = self._push_file
 
@@ -539,25 +464,9 @@ class SshConnection(SshConnectionBase):
         push(sftp, src, dst)
 
     def _pull_file(self, sftp, src, dst):
-        # Pulling a file into a folder will use the source basename
-        if os.path.isdir(dst):
-            dst = os.path.join(
-                dst,
-                os.path.basename(src),
-            )
-
-        with contextlib.suppress(FileNotFoundError):
-            os.remove(dst)
-
         sftp.get(src, dst, callback=self._get_progress_cb())
 
     def _pull_folder(self, sftp, src, dst):
-        with contextlib.suppress(FileNotFoundError):
-            try:
-                shutil.rmtree(dst)
-            except OSError:
-                os.remove(dst)
-
         os.makedirs(dst)
         for fileattr in sftp.listdir_attr(src):
             filename = fileattr.filename
@@ -625,6 +534,8 @@ class SshConnection(SshConnectionBase):
         try:
             with _handle_paramiko_exceptions(command):
                 exit_code, output = self._execute(command, timeout, as_root, strip_colors)
+        except TargetCalledProcessError:
+            raise
         except TargetStableError as e:
             if will_succeed:
                 raise TargetTransientError(e)
@@ -632,8 +543,13 @@ class SshConnection(SshConnectionBase):
                 raise
         else:
             if check_exit_code and exit_code:
-                message = 'Got exit code {}\nfrom: {}\nOUTPUT: {}'
-                raise TargetStableError(message.format(exit_code, command, output))
+                cls = TargetTransientCalledProcessError if will_succeed else TargetStableCalledProcessError
+                raise cls(
+                    exit_code,
+                    command,
+                    output,
+                    None,
+                )
             return output
 
     def background(self, command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, as_root=False):
@@ -644,6 +560,7 @@ class SshConnection(SshConnectionBase):
         return bg_cmd
 
     def _background(self, command, stdout, stderr, as_root):
+        orig_command = command
         stdout, stderr, command = redirect_streams(stdout, stderr, command)
 
         command = "printf '%s\n' $$; exec sh -c {}".format(quote(command))
@@ -654,7 +571,7 @@ class SshConnection(SshConnectionBase):
             # Read are not buffered so we will always get the data as soon as
             # they arrive
             return (
-                channel.makefile_stdin(),
+                channel.makefile_stdin('w', 0),
                 channel.makefile(),
                 channel.makefile_stderr(),
             )
@@ -666,7 +583,20 @@ class SshConnection(SshConnectionBase):
             timeout=None,
             executor=executor,
         )
-        pid = int(stdout_in.readline())
+        pid = stdout_in.readline()
+        if not pid:
+            stderr = stderr_in.read()
+            if channel.exit_status_ready():
+                ret = channel.recv_exit_status()
+            else:
+                ret = 126
+            raise subprocess.CalledProcessError(
+                ret,
+                command,
+                b'',
+                stderr,
+            )
+        pid = int(pid)
 
         def create_out_stream(stream_in, stream_out):
             """
@@ -685,11 +615,11 @@ class SshConnection(SshConnectionBase):
                 w = os.fdopen(w, 'wb')
             # Turn a file descriptor into a file-like object
             elif isinstance(stream_out, int) and stream_out >= 0:
-                r = os.fdopen(stream_out, 'rb')
+                r = os.fdopen(stream_in, 'rb')
                 w = os.fdopen(stream_out, 'wb')
             # file-like object
             else:
-                r = stream_out
+                r = stream_in
                 w = stream_out
 
             return (r, w)
@@ -731,6 +661,7 @@ class SshConnection(SshConnectionBase):
             # Make sure the writing end are closed proper since we are not
             # going to write anything anymore
             for r, w in out_streams.values():
+                w.flush()
                 if r is not w and w is not None:
                     w.close()
 
@@ -760,6 +691,7 @@ class SshConnection(SshConnectionBase):
             stdout=out_streams['stdout'][0],
             stderr=out_streams['stderr'][0],
             redirect_thread=redirect_thread,
+            cmd=orig_command,
         )
 
     def _close(self):
@@ -854,6 +786,7 @@ class TelnetConnection(SshConnectionBase):
             strict_host_check=strict_host_check,
         )
 
+        _check_env()
         self.options = self._get_default_options()
 
         self.lock = threading.Lock()
@@ -864,6 +797,63 @@ class TelnetConnection(SshConnectionBase):
         self.conn = telnet_get_shell(host, username, password, port, timeout, original_prompt)
         atexit.register(self.close)
 
+    def fmt_remote_path(self, path):
+        return '{}@{}:{}'.format(self.username, self.host, path)
+
+    def _get_default_options(self):
+        if self.strict_host_check:
+            options = {
+                'StrictHostKeyChecking': 'yes',
+            }
+        else:
+            options = {
+                'StrictHostKeyChecking': 'no',
+                'UserKnownHostsFile': '/dev/null',
+            }
+        return options
+
+    def push(self, sources, dest, timeout=30):
+        # Quote the destination as SCP would apply globbing too
+        dest = self.fmt_remote_path(quote(dest))
+        paths = list(sources) + [dest]
+        return self._scp(paths, timeout)
+
+    def pull(self, sources, dest, timeout=30):
+        # First level of escaping for the remote shell
+        sources = ' '.join(map(quote, sources))
+        # All the sources are merged into one scp parameter
+        sources = self.fmt_remote_path(sources)
+        paths = [sources, dest]
+        self._scp(paths, timeout)
+
+    def _scp(self, paths, timeout=30):
+        # NOTE: the version of scp in Ubuntu 12.04 occasionally (and bizarrely)
+        # fails to connect to a device if port is explicitly specified using -P
+        # option, even if it is the default port, 22. To minimize this problem,
+        # only specify -P for scp if the port is *not* the default.
+        port_string = '-P {}'.format(quote(str(self.port))) if (self.port and self.port != 22) else ''
+        keyfile_string = '-i {}'.format(quote(self.keyfile)) if self.keyfile else ''
+        options = " ".join(["-o {}={}".format(key, val)
+                            for key, val in self.options.items()])
+        paths = ' '.join(map(quote, paths))
+        command = '{} {} -r {} {} {}'.format(scp,
+                                                options,
+                                                keyfile_string,
+                                                port_string,
+                                                paths)
+        command_redacted = command
+        logger.debug(command)
+        if self.password:
+            command, command_redacted = _give_password(self.password, command)
+        try:
+            check_output(command, timeout=timeout, shell=True)
+        except subprocess.CalledProcessError as e:
+            raise_from(HostError("Failed to copy file with '{}'. Output:\n{}".format(
+                command_redacted, e.output)), None)
+        except TimeoutError as e:
+            raise TimeoutError(command_redacted, e.output)
+
+
     def execute(self, command, timeout=None, check_exit_code=True,
                 as_root=False, strip_colors=True, will_succeed=False): #pylint: disable=unused-argument
         if command == '':
@@ -883,16 +873,23 @@ class TelnetConnection(SshConnectionBase):
                 if check_exit_code:
                     try:
                         exit_code = int(exit_code_text)
-                        if exit_code:
-                            message = 'Got exit code {}\nfrom: {}\nOUTPUT: {}'
-                            raise TargetStableError(message.format(exit_code, command, output))
                     except (ValueError, IndexError):
-                        logger.warning(
+                        raise ValueError(
                             'Could not get exit code for "{}",\ngot: "{}"'\
                             .format(command, exit_code_text))
+                    if exit_code:
+                        cls = TargetTransientCalledProcessError if will_succeed else TargetStableCalledProcessError
+                        raise cls(
+                            exit_code,
+                            command,
+                            output,
+                            None,
+                        )
                 return output
         except EOF:
             raise TargetNotRespondingError('Connection lost.')
+        except TargetCalledProcessError:
+            raise
         except TargetStableError as e:
             if will_succeed:
                 raise TargetTransientError(e)
@@ -1102,9 +1099,6 @@ class Gem5Connection(TelnetConnection):
             # We need to copy the file to copy to the temporary directory
             self._move_to_temp_dir(source)
 
-            # Dest in gem5 world is a file rather than directory
-            if os.path.basename(dest) != filename:
-                dest = os.path.join(dest, filename)
             # Back to the gem5 world
             filename = quote(self.gem5_input_dir + filename)
             self._gem5_shell("ls -al {}".format(filename))
@@ -1173,7 +1167,10 @@ class Gem5Connection(TelnetConnection):
         try:
             output = self._gem5_shell(command,
                                       check_exit_code=check_exit_code,
-                                      as_root=as_root)
+                                      as_root=as_root,
+                                      will_succeed=will_succeed)
+        except TargetCalledProcessError:
+            raise
         except TargetStableError as e:
             if will_succeed:
                 raise TargetTransientError(e)
@@ -1407,7 +1404,7 @@ class Gem5Connection(TelnetConnection):
             raise TargetStableError('Path to m5 binary on simulated system  is not set!')
         self._gem5_shell('{} {}'.format(self.m5_path, command))
 
-    def _gem5_shell(self, command, as_root=False, timeout=None, check_exit_code=True, sync=True):  # pylint: disable=R0912
+    def _gem5_shell(self, command, as_root=False, timeout=None, check_exit_code=True, sync=True, will_succeed=False):  # pylint: disable=R0912
         """
         Execute a command in the gem5 shell
 
@@ -1472,11 +1469,17 @@ class Gem5Connection(TelnetConnection):
                                              sync=False)
             try:
                 exit_code = int(exit_code_text.split()[0])
-                if exit_code:
-                    message = 'Got exit code {}\nfrom: {}\nOUTPUT: {}'
-                    raise TargetStableError(message.format(exit_code, command, output))
             except (ValueError, IndexError):
-                gem5_logger.warning('Could not get exit code for "{}",\ngot: "{}"'.format(command, exit_code_text))
+                raise ValueError('Could not get exit code for "{}",\ngot: "{}"'.format(command, exit_code_text))
+            else:
+                if exit_code:
+                    cls = TragetTransientCalledProcessError if will_succeed else TargetStableCalledProcessError
+                    raise cls(
+                        exit_code,
+                        command,
+                        output,
+                        None,
+                    )
 
         return output
 
